﻿(*

LockFreePrimitives

Description:
 Lock-Free Primitives Collection

Homepage:
 http://www.idescriptor.com/RingBuffer/LockFreePrimitives.pas

Mail:
 gorazd.jernejc@gmail.com

Copyright (c) 2009, Gorazd Jernejc
All rights reserved.

License:
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
 * The name of the Gorazd Jernejc may not be used to endorse or promote products
   derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 COMPANIES WHOSE NAME STARTS WITH LETTERS 'DANF' MAY NOT USE THIS SOFTWARE!

History:
 Author            : GJ
 Creation date     : 4. August 2009
 Last modification : 20. August 2009
 Version           : 1.02

 1.00: 4. August 2009
 First official release.

 1.01: 17. August 2009
 Remark: Atomic128BitMove must by 16-byte aligned.

 1.02: 20. August 2009
 Added: GetRingBufferDistance

*)

unit LockFreePrimitives;

interface

type
  PReferencedPtr = ^TReferencedPtr;
  TReferencedPtr = record
    Link     : pointer;
    Reference: cardinal;
  end; { TReferencedPtr }

  PReferencedPtrBuffer = ^TReferencedPtrBuffer;
  TReferencedPtrBuffer = array [0..MaxInt shr 4] of TReferencedPtr;

  PRingBuffer = ^TRingBuffer;
  TRingBuffer  = record
    Left       : TReferencedPtr;
    Right      : TReferencedPtr;
    Length     : integer;
    EndBuffer  : pointer;
    Buffer     : record end;
  end; { TgjRingBuffer }

procedure Atomic128BitMove(var Source, Destination);
function  CAS32(const Old, New: cardinal; var Destination): boolean; overload;
function  CAS32(const Old, New: pointer; var Destination): boolean; overload;
function  CAS64(const OldData, OldReference, NewData, NewReference: cardinal;
  var Destination): boolean; overload;
function  CAS64(const OldData: pointer; const OldReference: cardinal;
  const NewData: pointer; const NewReference: cardinal;  var Destination): boolean; overload;
function  GetCPUTimeStamp: Int64;
function  GetRingBufferDistance(var RB: TRingBuffer): integer;
function  GetThreadId: cardinal;
procedure SFence;
function IsSSE2: boolean;

implementation

{ Intel Atomic functions support }

procedure Atomic128BitMove(var Source, Destination);
//Move 16 bytes atomicly from Source 16-byte aligned to Destination!
asm
  movdqa  xmm0, dqword [Source]
  movdqu  dqword [Destination], xmm0
end;

function CAS32(const Old, New: cardinal; var Destination): boolean;
//32 bit compare and swap
asm
  lock cmpxchg dword ptr [Destination], New
  setz  al
end; { CAS32 }

function CAS32(const Old, New: pointer; var Destination): boolean;
//32 bit compare and swap
asm
  lock cmpxchg dword ptr [Destination], New
  setz  al
end; { CAS32 }

function CAS64(const OldData, OldReference, NewData, NewReference: cardinal;
  var Destination): boolean;
//64 bit compare and swap
asm
  push  edi
  push  ebx
  mov   ebx, NewData
  mov   ecx, NewReference
  mov   edi, Destination
  lock cmpxchg8b qword ptr [edi]
  setz  al
  pop   ebx
  pop   edi
end; { CAS64 }

function CAS64(const OldData: pointer; const OldReference: cardinal;
  const NewData: pointer; const NewReference: cardinal;  var Destination): boolean;
//64 bit compare and swap
asm
  push  edi
  push  ebx
  mov   ebx, NewData
  mov   ecx, NewReference
  mov   edi, Destination
  lock cmpxchg8b qword ptr [edi]
  setz  al
  pop   ebx
  pop   edi
end; { CAS64 }

function GetCPUTimeStamp: Int64;
//Return current CPU tick
asm
  rdtsc
end; { GetTimeStamp }

function GetRingBufferDistance(var RB: TRingBuffer): integer;
//Source must be 16-byte aligned!
asm
  lea     edx, esp - $18                        //use free stack as temp memory
  and     edx, -$10                             //ensure 16-byte aligment
  movdqa  xmm0, dqword [RB]                     //128 bit atomic move to xmm0
  movdqa  dqword [edx], xmm0                    //128 bit atomic move to temp memory
  mov     ecx, TRingBuffer([edx]).Right.Link    //calcolate distance Right - Left
  sub     ecx, TRingBuffer([edx]).Left.Link
  mov     edx, ecx
  not     ecx
  shr     ecx, 3
  mov     eax, [RB].Length                      //eax := Length
  sub     eax, ecx
  shr     edx, 3
  bt      edx, 28
  cmovnc  eax, edx
end; { GetRingBufferDistance }


function GetThreadId: cardinal;
//Return ThreadId
asm
  mov   eax, fs:[$18]                   //eax := thread information block
  mov   eax, [eax + $24]                //eax := thread id
end; { GetThreadId }

procedure SFence;
//Perform memory save fence
asm
  sfence
end; { Sfence }

function IsSSE2: boolean;
//Return true if support SSE2
asm
  push  ebx
  mov   eax, 1
  cpuid
  bt    edx, 26
  setc  al
  pop   ebx
end; { IsSSE2 }

end.

