(*

RingBuffer

Description:
 Lock-free multy thread read/write conflict optimized RingBuffer, Queue and Stack.

Homepage:
 http://www.idescriptor.com/RingBuffer/RingBuffer.htm

Mail:
 gorazd.jernejc@gmail.com

Copyright (c) 2009, Gorazd Jernejc
All rights reserved.

License:
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
 * The name of the Gorazd Jernejc may not be used to endorse or promote products
   derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 COMPANIES WHOSE NAME STARTS WITH LETTERS 'DANF' MAY NOT USE THIS SOFTWARE!

History:
 Author            : GJ
 Creation date     : 4. August 2009
 Last modification : 27. August 2009
 Version           : 1.03

 1.00: 4. August 2009
 First official release.

 1.01: 17. August 2009
 TgjRingBuffer.Initialize: Ensured 16-byte buffer alignment.
 Added SSE2 test in initialization section.

 1.02: 20. August 2009
 TgjRingBuffer.Length is stored now in TRingBuffer record
 TgjRingBuffer.Count: Now call faster GetRingBufferDistance

 1.03: 27. August 2009
 Functions InsertLeft, InsertRight, RemoveLeft and RemoveRight now return the
 current position in ring buffer. If buffer is full or empty return 0.

*)

unit RingBuffer;

{$R-,O+,A8}

interface

uses
  LockFreePrimitives;

type
  TgjRingBuffer = class(TObject)
  strict private
    rbRingBuffer       : PRingBuffer;
    rbRingBufferMemPtr : pointer;
    class var LoopTicks: cardinal;
    procedure MeasureExecutionTimes;
  protected
    procedure Empty; inline;
    function  GetLength: integer; inline;
  public
    destructor Destroy; override;
    function  Count: integer; inline;
    procedure Initialize(const Length: integer);
    function  InsertLeft(const link: pointer): integer; overload; inline;
    function  InsertRight(const link: pointer): integer; inline;
    function  RemoveLeft(out link: pointer): integer; inline;
    function  RemoveRight(var link: pointer): integer; inline;
    property  Length: integer read GetLength;
  end; { TgjRingBuffer }

  TgjQueue = class(TgjRingBuffer)
  public
    function Dequeue(var link: pointer): integer; inline;
    function Enqueue(const link: pointer): integer; inline;
  end; { TgjQueue }

  TgjStack = class(TgjRingBuffer)
  public
    function Pop(var link: pointer): integer; inline;
    function Push(const link: pointer): integer; inline;
  end; { TgjStack }

implementation

uses
  SysUtils;

{ TgjRingBuffer }

function TgjRingBuffer.Count: integer;
begin
  Result := GetRingBufferDistance(rbRingBuffer^);
end; { TgjRingBuffer.Count }

destructor TgjRingBuffer.Destroy;
begin
  FreeMem(rbRingBufferMemPtr);
  inherited;
end; { TgjRingBuffer.Destroy }

procedure TgjRingBuffer.Empty;
begin
  with rbRingBuffer^ do begin
    Left.Link := @Buffer;
    Right.Link := Left.Link;
    EndBuffer := @PReferencedPtrBuffer(@Buffer)^[Length];
  end;
end; { TgjRingBuffer.Empty }

function TgjRingBuffer.GetLength: integer;
begin
  result := rbRingBuffer.Length;
end; { TgjRingBuffer.Length }

procedure TgjRingBuffer.Initialize(const Length: integer);
begin
  Assert(SizeOf(cardinal) = SizeOf(pointer));
  Assert(Length > 0);
  // allocate RingBuffer
  Assert (rbRingBuffer = nil,'OmniRingBuffer: Error multiple Initialize');
  GetMem(rbRingBufferMemPtr, SizeOf(TReferencedPtr) * (Length + 1) + SizeOf(TRingBuffer) + $0C);
  //Ensure 16-byte alignment
  rbRingBuffer := pointer((cardinal(rbRingBufferMemPtr) + $0C) AND NOT $0F);
  //Set Length
  rbRingBuffer.Length := Length;
  Empty;
  MeasureExecutionTimes;
end; { TgjRingBuffer.Initialize }

function TgjRingBuffer.InsertLeft(const link: pointer): integer;
var
  AtStartReference: cardinal;
  CPUTimeStamp    : int64;
  CurrentLeft     : pointer;
  CurrentReference: cardinal;
  NewLeft         : PReferencedPtr;
  Reference       : cardinal;
label
  TryAgain;
begin
  Reference := GetThreadId + 1;                 //Reference.bit0 := 1
  with rbRingBuffer^ do begin
TryAgain:
    //Set Left.Reference with respect to all other cores :)
    CPUTimeStamp := GetCPUTimeStamp + LoopTicks;
    AtStartReference := Left.Reference OR 1;    //Reference.bit0 := 1
    repeat
      CurrentReference := Left.Reference;
    until (CurrentReference AND 1 = 0)or (GetCPUTimeStamp - CPUTimeStamp > 0);
    //No threads present in ring buffer or current thread timeout
    if ((CurrentReference AND 1 <> 0) and (AtStartReference <> CurrentReference)) or
      not CAS32(CurrentReference, Reference, Left.Reference) then
      goto TryAgain;
    //Calculate RingBuffer NewLeft address
    CurrentLeft := Left.Link;
    NewLeft := pointer(cardinal(CurrentLeft) - SizeOf(TReferencedPtr));
    if cardinal(NewLeft) < cardinal(@Buffer) then
      NewLeft := EndBuffer;
    //Calcolate distance
    result := integer(Right.Link) - Integer(NewLeft);
    //Check buffer full
    if result = 0 then                  //Clear Reference if task still own reference
      if CAS32(Reference, 0, Left.Reference) then
        Exit else
        goto TryAgain;
    //Set NewLeft.Reference
    NewLeft^.Reference := Reference;
    SFence;
    //Try to set link and try to exchange NewLeft and clear Reference if task own reference
    if (Reference <> Left.Reference) or
      not CAS64(NewLeft^.Link, Reference, link, Reference, NewLeft^) or
      not CAS64(CurrentLeft, Reference, NewLeft, 0, Left) then
      goto TryAgain;
    //Calcolate result
    if result < 0 then
      result := Length - integer(cardinal(not Result) div SizeOf(TReferencedPtr)) else
      result := cardinal(result) div SizeOf(TReferencedPtr);
  end; //with
end; { TgjRingBuffer.InsertLeft }

function TgjRingBuffer.InsertRight(const link: pointer): integer;
var
  AtStartReference: cardinal;
  CPUTimeStamp    : int64;
  CurrentRight    : PReferencedPtr;
  CurrentReference: cardinal;
  NewRight        : PReferencedPtr;
  Reference       : cardinal;
label
  TryAgain;
begin
  Reference := GetThreadId + 1;                 //Reference.bit0 := 1
  with rbRingBuffer^ do begin
TryAgain:
    //Set Right.Reference with respect to all other cores :)
    CPUTimeStamp := GetCPUTimeStamp + LoopTicks;
    AtStartReference := Right.Reference OR 1;   //Reference.bit0 := 1
    repeat
      CurrentReference := Right.Reference;
    until (CurrentReference AND 1 = 0) or (GetCPUTimeStamp - CPUTimeStamp - LoopTicks > 0);
    //No threads present in ring buffer or current thread timeout
    if ((CurrentReference AND 1 <> 0) and (AtStartReference <> CurrentReference)) or
      not CAS32(CurrentReference, Reference, Right.Reference) then
      goto TryAgain;
    //Set CurrentRight.Reference
    CurrentRight := Right.Link;
    CurrentRight^.Reference := Reference;
    SFence;
    //Calculate ringBuffer next Right address
    NewRight := pointer(cardinal(CurrentRight) + SizeOf(TReferencedPtr));
    if cardinal(NewRight) > cardinal(EndBuffer) then
      NewRight := @Buffer;
    //Calcolate distance
    result := integer(NewRight) - Integer(Left.Link);
    //Check buffer full
    if result = 0 then                  //Clear Reference if task still own reference
      if CAS32(Reference, 0, Right.Reference) then
        Exit else
        goto TryAgain;
    //Try to Set link and try to exchange NewRight and clear Reference if task own reference
    if (Reference <> Right.Reference) or
      not CAS64(CurrentRight.Link, Reference, link, Reference, CurrentRight^) or
      not CAS64(CurrentRight, Reference, NewRight, 0, Right) then
      goto TryAgain;
    //Calcolate result
    if result < 0 then
      result := Length - integer(cardinal(not Result) div SizeOf(TReferencedPtr)) else
      result := cardinal(result) div SizeOf(TReferencedPtr);
  end; //with
end;

{ TgjRingBuffer.InsertRight }

const
  NumOfInitSamples = 10;

procedure TgjRingBuffer.MeasureExecutionTimes;
var
  Sample       : integer;
  TimeTestField: array [0..1]of array [1..NumOfInitSamples] of int64;

  function AddMinAndClear(Rutine, Count: cardinal): int64;
  var
    first    : boolean;
    n        : cardinal;
    Sample   : integer;
    Minimum  : integer;
  begin
    Result := 0;
    first := true;
    for n := 1 to Count + 1 do begin
      Minimum:= 1;
      for Sample:= 2 to NumOfInitSamples do
        if TimeTestField[Rutine, Sample] < TimeTestField[Rutine, Minimum] then
          Minimum := Sample;
      if first then
        first := false                          //Ignore first minimum
      else
        Inc(Result, TimeTestField[Rutine, Minimum]);
      TimeTestField[Rutine, Minimum] := MaxLongInt;
    end;
  end; { GetMinAndClear }

begin
  if LoopTicks = 0 then begin
    //Calculate rbLoopTicks and rbInsertTicks counter values depend on CPU speed!}
    LoopTicks := 1;
    for Sample := 1 to NumOfInitSamples do  begin
      //Measure InsertRight rutine delay
      TimeTestField[0, Sample] := GetCPUTimeStamp;
      InsertRight(nil);
      TimeTestField[0, Sample] := GetCPUTimeStamp - TimeTestField[0, Sample];
      //Measure GetCPUTimeStamp rutine delay
      TimeTestField[1, Sample] := GetCPUTimeStamp;
      TimeTestField[1, Sample] := GetCPUTimeStamp - TimeTestField[1, Sample];
    end;
    //Calculate 4 minimum average (except first one) for rbLoopTicks * 2
    LoopTicks := (AddMinAndClear(0, 4) - AddMinAndClear(1, 4)) * 4;
  end;
end; { TgjRingBuffer.MeasureExecutionTimes }

function TgjRingBuffer.RemoveLeft(out link: pointer): integer;
var                                                                  
  AtStartReference: cardinal;
  CPUTimeStamp    : int64;
  CurrentLeft     : PReferencedPtr;
  CurrentReference: cardinal;
  NewLeft         : PReferencedPtr;
  Reference       : cardinal;
label
  TryAgain;
begin
  Reference := GetThreadId + 1;                 //Reference.bit0 := 1
  with rbRingBuffer^ do begin
TryAgain:
    //Set Left.Reference with respect to all other cores :)
    CPUTimeStamp := GetCPUTimeStamp + LoopTicks;
    AtStartReference := Left.Reference OR 1;    //Reference.bit0 := 1
    repeat
      CurrentReference := Left.Reference;
    until (CurrentReference AND 1 = 0) or (GetCPUTimeStamp - CPUTimeStamp - LoopTicks > 0);
    //No threads present in ring buffer or current thread timeout
    if ((CurrentReference AND 1 <> 0) and (AtStartReference <> CurrentReference)) or
      not CAS32(CurrentReference, Reference, Left.Reference) then
      goto TryAgain;
    //Calcolate distance
    CurrentLeft := Left.Link;
    result := integer(Right.Link) - Integer(CurrentLeft);
    //Buffer empty test
    if result = 0 then
      if CAS32(Reference, 0, Left.Reference) then
      Exit else
      goto TryAgain;
    //Load target link
    link := PReferencedPtr(Left.Link).Link;
    //Calculate ringBuffer next Left address
    NewLeft := pointer(cardinal(CurrentLeft) + SizeOf(TReferencedPtr));
    if cardinal(NewLeft) > cardinal(EndBuffer) then
      NewLeft := @Buffer;
    //Try to exchange NewLeft and clear Reference if task own reference
    if not CAS64(CurrentLeft, Reference, NewLeft, 0, Left) then
      goto TryAgain;
    //Calcolate result
    if result < 0 then
      result := Length - integer(cardinal(not Result) div SizeOf(TReferencedPtr)) else
      result := cardinal(result) div SizeOf(TReferencedPtr);
  end; //with
end; { TgjRingBuffer.RemoveLeft }

function TgjRingBuffer.RemoveRight(var link: pointer): integer;
var
  AtStartReference: cardinal;
  CPUTimeStamp    : int64;
  CurrentRight    : PReferencedPtr;
  CurrentReference: cardinal;
  NewRight        : PReferencedPtr;
  Reference       : cardinal;
label
  TryAgain;
begin
  Reference := GetThreadId + 1;                 //Reference.bit0 := 1
  with rbRingBuffer^ do begin
TryAgain:
    //Set Right.Reference with respect to all other cores :)
    CPUTimeStamp := GetCPUTimeStamp + LoopTicks;
    AtStartReference := Right.Reference OR 1;   //Reference.bit0 := 1
    repeat
      CurrentReference := Right.Reference;
    until (CurrentReference AND 1 = 0) or (GetCPUTimeStamp - CPUTimeStamp > 0);
    //No threads present in ring buffer or current thread timeout
    if (CurrentReference AND 1 <> 0) and (AtStartReference <> CurrentReference) or
      not CAS32(CurrentReference, Reference, Right.Reference) then
      goto TryAgain;
    //Calculate ringBuffer NewRight address
    CurrentRight := Right.Link;
    NewRight := pointer(cardinal(CurrentRight) - SizeOf(TReferencedPtr));
    if cardinal(NewRight) < cardinal(@Buffer) then
      NewRight := EndBuffer;
    //Calcolate distance
    result := integer(CurrentRight) - Integer(Left.Link);
    //Buffer empty test
    if result = 0 then
      if CAS32(Reference, 0, Right.Reference) then
      Exit else
      goto TryAgain;
    //Load target link
    link := NewRight.Link;
    //Try to exchange NewRight and clear Reference if current task own reference
    if not CAS64(CurrentRight, Reference, NewRight, 0, Right) then
      goto TryAgain;
    //Calcolate result
    if result < 0 then
      result := Length - integer(cardinal(not Result) div SizeOf(TReferencedPtr)) else
      result := cardinal(result) div SizeOf(TReferencedPtr);
  end;  //with
end; { TgjRingBuffer.RemoveRight }

{ TgjQueue }

function TgjQueue.Dequeue(var link: pointer): integer;
begin
  result := inherited RemoveRight(link);
end; { TgjQueue.Dequeue }

function TgjQueue.Enqueue(const link: pointer): integer;
begin
  result := inherited InsertLeft(link);
end; { TgjQueue.Enqueue }

{ TgjStack }

function TgjStack.Pop(var link: pointer): integer;
begin
  result := inherited RemoveLeft(link);
end; { TgjStack.Pop }

function TgjStack.Push(const link: pointer): integer;
begin
  result := inherited InsertLeft(link);
end; { TgjStack.Push }

initialization
  if not IsSSE2 then
    raise Exception.Create('Error: CPU doesn''t support SSE2');
  // InitializeTimingInfo
  with TgjRingBuffer.Create do begin
    Initialize(NumOfInitSamples);
    Free;
  end;

end.
